/*
 *   This file is part of pls4snp.
 *
 *   pls4snp is free software; you can redistribute it and/or modify
 *   it under the terms of the GNU General Public License as published by
 *   the Free Software Foundation; either version 2 of the License, or
 *   (at your option) any later version.
 *
 *   pls4snp is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *   GNU General Public License for more details.
 *
 *   You should have received a copy of the GNU General Public License
 *   along with pls4snp; if not, write to the Free Software
 *   Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
 *
 *   Copyright 2013 Olivier Filangi
 *
 */
/*
 * command.cpp
 *
 *  Created on: 14 mars 2013
 *      Author: ofilangi
 */
#include "command.hpp"

#include <iostream>
#include <fstream>
#include <ctime>
#include <string>
using namespace std;

#include "common.hpp"
#include "convert.hpp"
#include "factoryData.hpp"
#include "dataGen.hpp"
#include "calculDataGen.hpp"
#include "pls.hpp"
#include "predict.hpp"


void Command::command_spls(const int nkeep,
		int ncomp,
		int maxiter,
		DT tol,
		const string & phenotypeFile_ap,
		const string & genotypeFile_ap,
		const string & binprefix,
		bool edc_present,
		bool edc_calcul,
		int codeFormatGeno,
		bool headerGeno,
		DT mafmin,
		bool removeMinMaf) {
	dataGen data;
	workPls work ;

	factoryData::buildDataGenPhenotype(phenotypeFile_ap,data,edc_present);
	factoryData::buildDataGenGenotype(genotypeFile_ap,data,codeFormatGeno,headerGeno);

	factoryData::reorganize_Y_Z(data);

	int * listSnp = new int[data.p];
	int nrem = 0;
	nrem = factoryData::check_minor_allel_frequency(data,mafmin,listSnp);

	if ( removeMinMaf && (nrem > 0)) {
		int * listOk = new int[data.p - nrem];
		factoryData::remove_minmaf_snp(data,listSnp,nrem,listOk);
		delete [] listOk ;
	} else {
		nrem = 0;
	}

	if ( edc_calcul ) CalculDataGen::scaleWithEDC(data);

	if ( nkeep > data.p ) {
		cerr << "Wrong value defined for spls option ["<< nkeep << ">"<< data.p << "]" << endl ;
		exit(1);
	}

	if ( nkeep <= 0 ) {
		cerr << "Wrong value defined for spls option (<= 0) " << endl ;
		exit(1);
	}

	cout << "[spls] keepX="<< nkeep << endl ;

	work.spls2(data,data.mapY.begin()->first,ncomp,maxiter,tol,nkeep);
	ofstream os;
	string it="spls_"+Convert::T_to_string<int>(nkeep)+".bin";
	if ( binprefix != "" ) {
		it = binprefix + "_" + it ;
	}
	os.open (it.c_str(), ios::binary );
	work.write(&os,nrem,listSnp);
	os.close();

	cout << "Use the binary file ["<< it << "] to valid , predict and compute VIP SNP effect." << endl ;

	delete[] listSnp;
}

void Command::command_pls(int ncomp,
		int maxiter,
		DT tol,
		const string & phenotypeFile_ap,
		const string & genotypeFile_ap,
		const string & binprefix,
		bool edc_present,
		bool edc_calcul,
		int codeFormatGeno,
		bool headerGeno,
		DT mafmin,
		bool removeMinMaf) {
	dataGen data;
	workPls work ;

	factoryData::buildDataGenPhenotype(phenotypeFile_ap,data,edc_present);
	factoryData::buildDataGenGenotype(genotypeFile_ap,data,codeFormatGeno,headerGeno);

	factoryData::reorganize_Y_Z(data);

	/* pour debug */
	//data.n = data.nphenoRecord ;

	int * listSnp = new int[data.p];
	int nrem = 0;

	nrem = factoryData::check_minor_allel_frequency(data,mafmin,listSnp);

	if ( removeMinMaf && (nrem > 0) ) {
		int * listOk = new int[data.p - nrem];
		factoryData::remove_minmaf_snp(data,listSnp,nrem,listOk);
		delete [] listOk ;
	} else {
		nrem = 0;
	}

	if ( edc_calcul ) CalculDataGen::scaleWithEDC(data);

	work.spls2(data,data.mapY.begin()->first,ncomp,maxiter,tol);

	ofstream os;
	string outFileBin = "pls.bin";
	if ( binprefix != "" ) {
		outFileBin = binprefix + "_" + outFileBin ;
	}
	os.open (outFileBin.c_str(), ios::binary );
	work.write(&os,nrem,listSnp);
	os.close();

	cout << "Use the binary file ["<< outFileBin << "] to valid , predict and compute VIP SNP effect !" << endl ;

	delete[] listSnp;
}

void Command::command_vip(const string & binary,int H,const string & out) {
	int * listSnp = NULL;
	int nrem=0;

	workPls work ;
	ifstream is;
	is.open (binary.c_str(), ios::binary );
	work.read(&is,nrem,&listSnp);
	is.close();

	if ( H > work.ncomp || (H < 1) ) {
		cerr << "can not generated for H="<< H << " NCOMP="<< work.ncomp << endl ;
		exit(1);
	}

	/* allocation */
	DT ** VIP = new DT*[H];
	for (int h=0;h<H;h++)
		VIP[h] = new DT[work.p];

	/* calcul des VIP */
	work.calcul_vip(VIP,H);

	/* Ecriture des VIP dans un fichier resultat */
	ofstream of ;
	of.open(out.c_str());
	for (int i=0;i<work.p;i++) {
		for (int h=0;h<H;h++) {
			of << VIP[h][i]<< " " ;
		}
		of << endl ;
	}
	of.close();
	/* deallocation */
	for (int h=0;h<H;h++)
		delete[] VIP[h];
	delete[] VIP;
}

void Command::command_valid(const string & phenotypeFile_val,
		const string & genotypeFile_val,
		const string & binary,
		bool edc_present,
		bool edc_calcul,
		int codeFormatGeno,
		bool headerGeno) {

	dataGen dataPredict;

	factoryData::buildDataGenPhenotype(phenotypeFile_val,dataPredict,edc_present);
	factoryData::buildDataGenGenotype(genotypeFile_val,dataPredict,codeFormatGeno,headerGeno);

	/* on garde les individus qui ont un phenotype et un genotype */
	factoryData::reorganize_Y_Z(dataPredict);

	int * listSnp = NULL;
	int nrem=0;

	workPls work ;

	ifstream is;
	is.open (binary.c_str(), ios::binary );
	work.read(&is,nrem,&listSnp);
	is.close();
	int * listOk = NULL;
	int pAll = dataPredict.p ;

	// verification number of marker
	if ( work.p != (dataPredict.p - nrem) ) {
		cerr << "binary file have been constructed with "<< work.p << " markers !" << endl ;
		cerr << "data genotype file ["<< genotypeFile_val<< "] contains "<< dataPredict.p << " markers with "<< nrem << " markers to removed" << endl ;
		exit(1);
	}

	if ( nrem > 0 ) {
		cout << "Marker index :" ;
		for (int i=0;i<nrem;i++)
			cout << listSnp[i]<<" ";
		cout << "are not used in the prediction" << endl ;
		listOk = new int[dataPredict.p - nrem];
		factoryData::remove_minmaf_snp(dataPredict,listSnp,nrem,listOk);
	} else {
		listOk = new int[work.p];
		for (int i=0;i<work.p;i++)
			listOk[i]=i;
	}

	solPls sol;
	sol.predict(work,dataPredict.Z,dataPredict.n,pAll,listOk);

	/* Calcul des correlations */
	for (map<string,DT*>::const_iterator it=dataPredict.mapY.begin();it!=dataPredict.mapY.end();it++) {
		string name = (*it).first ;
		cout << " == correlation between observed and predicted values of "<< name << " == " << endl ;
		DT maxp=-2;
		int hmax=0;
		for (int h=0;h<work.ncomp;h++) {
			DT p = 0;
			if (edc_calcul) {
				p = CalculDataGen::correlation_edc(sol.yhat[h],dataPredict,name);
			} else {
				map<string,DT*>::const_iterator it2 = dataPredict.mapY.find(name);
				p = CalculDataGen::correlation(sol.yhat[h],(*it2).second,dataPredict.n);
			}
			cout << "h="<< h+1  << " c="<< p << endl ;
			if ( p>maxp) {
				maxp=p;
				hmax=h;
			}
		}
		cout << "hmax="<< hmax+1 << " corr="<< maxp << endl ;
	}

	//sol.write_sol(dataPredict,binary.substr(0,binary.size()-4)+"_");
	delete[] listOk;
	delete[] listSnp;
}

void Command::command_predict(
		const string & genotypeFile_val,
		const string & binary,
		int H,
		int codeFormatGeno,
		bool headerGeno,
		const string & nameOutFile) {
	dataGen dataPredict;
	workPls work ;

	factoryData::buildDataGenGenotype(genotypeFile_val,dataPredict,codeFormatGeno,headerGeno);

	int * listSnp = NULL;
	int nrem=0;

	solPls sol;

	cout << "read binary file:"<< binary << endl ;
	ifstream is;
	is.open (binary.c_str(), ios::binary );
	work.read(&is,nrem,&listSnp);
	is.close();
	int * listOk = NULL;
	int pAll = dataPredict.p ;

	// verification number of marker
	if ( work.p != (dataPredict.p - nrem) ) {
		cerr << "binary file have been constructed with "<< work.p << " markers !" << endl ;
		cerr << "data genotype file ["<< genotypeFile_val<< "] contains "<< dataPredict.p << " markers with "<< nrem << " markers to removed" << endl ;
		exit(1);
	}

	if ( nrem > 0 ) {
		cout << "Marker index :" ;
		for (int i=0;i<nrem;i++)
			cout << listSnp[i]<<" ";
		cout << "are not used in the prediction" << endl ;
		listOk = new int[dataPredict.p - nrem];
		factoryData::remove_minmaf_snp(dataPredict,listSnp,nrem,listOk);
	} else {
		listOk = new int[work.p];
		for (int i=0;i<work.p;i++)
			listOk[i]=i;
	}

	sol.predict(work,dataPredict.Z,dataPredict.ngenoRecord,pAll,listOk,H);
	sol.write_sol(dataPredict,nameOutFile,H-1);

	delete[] listOk;
	delete[] listSnp;
}

