/*
 *   This file is part of pls4snp.
 *
 *   pls4snp is free software; you can redistribute it and/or modify
 *   it under the terms of the GNU General Public License as published by
 *   the Free Software Foundation; either version 2 of the License, or
 *   (at your option) any later version.
 *
 *   pls4snp is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *   GNU General Public License for more details.
 *
 *   You should have received a copy of the GNU General Public License
 *   along with pls4snp; if not, write to the Free Software
 *   Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
 *
 *   Copyright 2013 Olivier Filangi
 *
 */
#include "common.hpp"
#include "pls.hpp"

#include "calculDataGen.hpp"

#include <math.h>
#include <string>
#include <iostream>
#include <fstream>
using namespace std;

workPls::workPls()
{
	n = 0;
	p = 0;
	ncomp = 0;
	meanY = 0;
	sigmaY = 1;
	meanZ= NULL ;
	sigmaZ = NULL ;
	t = NULL;
	u = NULL;
	a = NULL;
	b = NULL;
	c = NULL;
	Y = NULL ;
}

void workPls::release() {
	for (int h=0; h<ncomp; h++)
	{
		if (t) delete[] t[h];
		if (u) delete[] u[h];
		if (a) delete[] a[h];
		if (c) delete[] c[h];
	}

	if (t) delete[] t;
	if (u) delete[] u;
	if (a) delete[] a;
	if (b) delete[] b;
	if (c) delete[] c;

	if ( meanZ ) delete[] meanZ;
	if ( sigmaZ ) delete[] sigmaZ;

	if ( Y ) delete[] Y;
}

workPls::~workPls()
{
	release();
}

/*
 * Interface to call pls/spls from a programme C/C++/fortran
 *
 * Purpose:
 *
 * Parameters:
 *   [in]  n          : specifies number of row of Z and Y
 *   [in]  p          : specifies number of SNP (column) of Z
 *   [in]  Z[n,p]     : incidence matrix
 *   [in]  Y[n]       : performance vector
 *   [in]  edc_pres   : 0 if EDC values are available otherwise >0.
 *   [in]  EDC[n]     : EDC vector
 *   [in]  ncomp      : specifies number of latent variable to build
 *   [in]  maxiter    : maximum number of iteration for convergence
 *   [in]  tol        : criteria for convergence
 *   [in]  keepX      : number of SNP to keep (Sparse PLS option), -1 to perform a PLS.
 *   [out] meanY      : mean of Y
 *   [out] sigmY      : standard deviation of Y
 *   [out] meanZ[p]   : vector of mean of each SNP column
 *   [out] sigmZ[p]   : vector of standard deviation of each SNP column
 *   [out] t[n,ncomp] :
 *   [out] u[n,ncomp] :
 *   [out] a[p,ncomp] :
 *   [out] b[ncomp]   :
 *   [out] c[p,ncomp] :
 */
extern "C" void spls_(int * n, int * p, DT * Z, DT * Y, int * edc_pres, DT * EDC, int * ncomp, int * maxiter, DT * tol,int * keepX,
		DT * meanY, DT * sigmY, DT * meanZ, DT * sigmZ, DT * t, DT * u, DT * a, DT * b, DT * c ) {
	/* initialisation d'une variable DataGen */
	dataGen data;

	data.Z = new DT*[*n];
	DT * nY = new DT[*n];
	DT * nEDC = NULL ;

	if ( *edc_pres == 0 ) {
		nEDC = new DT[*n];
	}

	for (int i=0;i<*n;i++) {
		nY[i] = Y[i];
		if ( *edc_pres == 0 ) nEDC[i] = EDC[i];

		data.Z[i] = new DT[*p];
		for ( int j=0;j<*p;j++)
			data.Z[i][j] = Z[i+j*(*n)];
	}

	data.mapY["pheno"]=nY;
	if ( *edc_pres == 0 )
		data.mapEDC["pheno"]=nEDC;

	data.n = *n;
	data.p = *p;

	workPls work;

	/* PLS or S-PLS */

	work.spls2(data,"pheno",*ncomp,*maxiter,*tol,*keepX);

	for (int i=0;i<data.n;i++) {
		for (int j=0;j<*ncomp;j++) {
			t[i+j*(data.n)] = work.t[j][i];
			u[i+j*(data.n)] = work.u[j][i];
		}
	}
	for (int i=0;i<data.p;i++) {
		for (int j=0;j<*ncomp;j++) {
			a[i+j*(data.p)] = work.a[j][i];
			c[i+j*(data.p)] = work.c[j][i];
		}
		meanZ[i] = work.meanZ[i];
		sigmZ[i] = work.sigmaZ[i];
	}
	for (int j=0;j<*ncomp;j++) {
		b[j] = work.b[j];
	}

	*meanY = work.meanY ;
	*sigmY = work.sigmaY ;

	/* free memory with destructeur of dataGen and workPls */
}

/**
X : genotypes
Y : phenotypes
p : number of markers
n : number of animals
ncomp : number of phenotype to considere
iter : max iteration algorithm
tol : tolerance algorithm
keepX : number of SNP

- Attention les tableaux X, Y sont modifiés en sortie de la procedure
 **/
void workPls::spls2(dataGen & data, string var,int ncomp, int maxiter, DT tol,int keepX)
{
	bool spls = (keepX > 0)&&(keepX <data.p);
	int nx = data.p - keepX;

	DT * Y = data.mapY[var];

	int incOne=1;

	/* allocate solution */
	this->n=data.n;
	this->p=data.p;
	this->ncomp=ncomp;

	DT * columnIsNa = new DT[p];
	this->t = new DT*[ncomp];
	this->u = new DT*[ncomp];
	this->a = new DT*[ncomp];
	this->b = new DT[ncomp];
	this->c = new DT*[ncomp];

	this->meanZ = new DT[p];
	this->sigmaZ = new DT[p];
	DT * bufA = new DT[p];

	cout << "[pls] start "<< " n:"<< n <<" p:"<< p << " ncomp="<< ncomp << " keepX="<<keepX << endl ;
	//standart core Y
	standart1(Y,n,this->meanY,this->sigmaY);
	//standart core X by column

	/* Sauvegarde de Y dpour calculer plus tard les VIP */
	this->Y = new DT[n];
	for (int i=0;i<n;i++)
		this->Y[i] = Y[i];

	standart_Z(data.Z,n,p,columnIsNa,this->meanZ,this->sigmaZ) ;

	cout << "H=" ;
	for (int h=0; h<ncomp; h++)
	{
		this->t[h] = new DT[n];
		this->u[h] = new DT[n];
		this->a[h] = new DT[p];
		this->c[h] = new DT[p];

		int iter=0;
		DT a_old[p] ;
		DT b_old=0.0;

		for (int j=0; j<p; j++)
		{
			a_old[j] = 0.0 ;
		}

		// init u
		for (int i=0; i<n; i++)
		{
			this->u[h][i]=Y[i];
		}
		DT curTol=0.0;
		DT crosst=0.0;

		this->b[h] = 0;

		do
		{
			iter++;
			b_old=this->b[h];

			DT crossu;
			/*
		  R : drop(crossprod(u))
			 */
			crossu = nrm2(&data.n,this->u[h],&incOne);
			crossu *=crossu;

			//DT invcrossu = 1.0 / crossu*crossu;

			/*
		  R : a = t(X.temp) %*% u / drop(crossprod(u))
			 */
			step_calcul_tx_y_divcross(this->a[h],data.Z,this->u[h],crossu);

			//			cout << "crossu="<< crossu << endl ;
			//			cout << "sqrtcrossa="<< dnrm2_(&data.p,this->a[h],&incOne) << endl;

			if ( spls ) {
				/* Soft-thresholding penalty function => Sparse PLS*/
				step_soft_thres_penalty(this->a[h],bufA,nx);
			}

			/*
		 R :  a = a / drop(sqrt(crossprod(a)))
		      Always 1 ?
			 */
			DT crossa=update_a_crossa(this->a[h]);

			/*
		 R : t = X.temp %*% a / drop(crossprod(a))
			 */
			step_calcul_t(this->t[h],data.Z,this->a[h],crossa);

			/*
		 R: b = t(Y.temp) %*% t / drop(crossprod(t))
			 */
			step_calcul_b(this->b[h],crosst,Y, this->t[h]);
			/*
		 R : u = Y.temp %*% b / drop(crossprod(b))
			 */
			//			for (int i=0; i<n; i++)
			//			{
			//				this->u[h][i] = Y[i]/this->b[h];
			//			}
			step_calcul_u(this->u[h],Y,this->b[h]);

			curTol=step_calcul_curtol(a_old,this->a[h]);
		}
		while ( ! ((curTol < tol && ((this->b[h]-b_old)*(this->b[h]-b_old) < tol) ) || iter>=maxiter) );
		//cout << "iter="<< iter << " b="<< this->b[h] << endl;

		//DT cross;
		/*
	 R : c = t(X.temp) %*% t / drop(crossprod(t))
		 */

		step_calcul_tx_y_divcross(this->c[h],data.Z,this->t[h],crosst);
		/*
	 R : X.na = X.na - t %*% t(c)
		 */

		step_update_z(data.Z,this->t[h],this->c[h],columnIsNa);
		/*
	 R : d = t(Y.temp) %*% t /drop((t(t) %*% t))
		 */
		DT d = step_calcul_d(Y,this->t[h],crosst);

		//cout << "d="<< d <<endl ;
		/*
	 R :  Y.na = Y.na - t %*% t(d)
		 */
		step_update_y(Y,this->t[h],d);
		cout << h+1<<" " ;
		cout.flush();
	}
	cout << endl ;
	delete[] columnIsNa;
	delete[] bufA;
}

/*
 * Interface to call pls/spls from a programme C/C++/fortran
 *
 * Purpose:
 *
 * Parameters:
 *   [in]  ncomp         : specifies number of latent variable associated with t and a
 *   [in]  hcomp         : specifies number of latent variable to include
 *   [in]  n             : specifies number of row of Y
 *   [in]  p             : specifies number of SNP
 *   [in]  Y[n]          : performance vector
 *   [in]  t[n,ncomp]    :
 *   [in]  a[p,ncomp   ] :
 *   [out] VIP[p,ncomp]  : VIP results
 */

extern "C" void  calcul_vip_(int *ncomp,int * hcomp,int *n,int *p, DT* Y, DT * t,DT *a,DT *VIP) {

	DT ** VIP_int = NULL;
	workPls work ;
	work.Y = Y;
	work.n = *n;
	work.p = *p;

	work.t = new DT*[*hcomp];
	work.a = new DT*[*hcomp];
	VIP_int = new DT*[*hcomp];

	for (int h=0;h<*hcomp;h++) {
		work.t[h] = new DT[work.n];
		work.a[h] = new DT[work.p];
		VIP_int[h] = new DT[work.p];
	}
	for (int h=0;h<*hcomp;h++) {
		for (int i=0;i<work.n;i++)
			work.t[h][i] = t[i+h*(work.n)];
		for (int i=0;i<work.p;i++)
			work.a[h][i] = a[i+h*(work.n)];
	}

	work.calcul_vip(VIP_int,*hcomp);

	/* results */
	for (int h=0;h<*hcomp;h++) {
		for (int i=0;i<work.p;i++) {
			VIP[i+h*(work.p)] = VIP_int[h][i];
		}
	}

	for (int h=0;h<*hcomp;h++) {
		delete[] work.t[h] ;
		delete[] work.a[h] ;
		delete[] VIP_int[h] ;
	}

	delete[] work.t ;
	delete[] work.a ;
	delete[] VIP_int ;

	work.t = NULL ;
	work.a = NULL ;
	/* do not deallocate Y...*/
	work.Y = NULL;
	work.n=0;
	work.p=0;
}

void  workPls::calcul_vip(DT **VIP,int Hcomp) {

	DT cor[Hcomp];


	for (int h=0;h<Hcomp;h++) {
		cor[h] = CalculDataGen::correlation(Y,t[h],n);
		cor[h] *= cor[h] ;
	}

	for (int i=0;i<p;i++)
		VIP[0][i] = sqrt(a[0][i]*a[0][i]*p);

	DT sumc = cor[0];

	for (int h=1;h<Hcomp;h++) {
		for (int i=0;i<p;i++) {
			VIP[h][i] = 0;
		}

		for (int ih=0;ih<=h;ih++) {
			for (int i=0;i<p;i++) {
				VIP[h][i] += cor[ih]*a[ih][i]*a[ih][i];
			}
		}
		sumc += cor[h] ;
		DT scale = p / sumc ;

		for (int i=0;i<p;i++) {
			VIP[h][i] = sqrt(VIP[h][i]*scale);
		}
	}
}


void workPls::step_calcul_tx_y_divcross(DT *a, DT ** Z,const DT * u,const DT & crossu) const {

	DT invcrossu = 1.0 / crossu;
	{
		for (int i=0; i<this->p; i++)
		{
			a[i] = 0.0 ;
		}

		for (int i=0; i<this->n; i++)
		{
			for (int j=0; j<this->p; j++) {
				a[j] += Z[i][j] * u[i];
			}
		}

		for (int i=0; i<this->p; i++)
		{
			a[i] *= invcrossu ;
		}
	}
}

static int compare (void const *a, void const *b)
{
	DT const *pa = (DT const*) a;
	DT const *pb = (DT const*) b;

	DT diff = *pa - *pb;
	if ( diff < 0 ) return -1 ;
	if ( diff == 0 ) return 0;
	return 1;
}


void workPls::step_soft_thres_penalty(DT *a,DT * bufA, int nx) {
	/* copy A to get the order */
	for (int i=0;i<p;i++)
		bufA[i] = fabs(a[i]);

	qsort(bufA,p,sizeof(DT),compare);

	DT pivot = fabs(bufA[nx]);

	for (int i=0;i<p;i++) {
		if ( fabs(a[i]) > pivot ) {
			if ( a[i] > 0 ) {
				a[i] = (fabs(a[i]) - pivot);
			} else {
				a[i] = - (fabs(a[i]) - pivot);
			}
		} else {
			a[i] = 0;
		}
	}
}

DT workPls::update_a_crossa(DT *a) {

	int incOne = 1;
	DT sqrtcrossa =  nrm2(&(this->p),a,&incOne);

	DT crossa=0;
	for (int j=0; j<this->p; j++)
	{
		a[j] = a[j] / sqrtcrossa;
		crossa += a[j]*a[j];
	}
	return crossa;
}

void workPls::step_calcul_t(DT *t, DT ** Z,const DT * a,const DT & crossa) const {
	DT invcrossa = 1.0 / crossa;

	for (int i=0; i<this->n; i++)
	{
		t[i]=0.0;
		for (int j=0; j<this->p; j++)
		{
			t[i] += Z[i][j] * a[j];
		}
		t[i] *= invcrossa ;
	}
}
void workPls::step_calcul_b(DT &b,DT &crosst,DT * Y, DT *t) const {
	crosst=0.0;
	b=0.0;
	//crossprod t
	for (int i=0; i<this->n; i++)
	{
		crosst += t[i]*t[i];
		b += Y[i]*t[i];
	}

	b /= crosst;
}

void workPls::step_calcul_u(DT *u, DT *Y,DT b) const {

	DT invb=1.0/b;
	for (int i=0; i<this->n; i++)
	{
		u[i] = Y[i]*invb;
	}
}

DT  workPls::step_calcul_curtol(DT *a_old,DT *a) const {
	DT curTol=0.0;
	for (int j=0; j<this->p; j++)
	{
		float v = (a[j] - a_old[j]);
		curTol = curTol + v*v;
		a_old[j] = a[j];
	}
	return curTol;
}

void workPls::step_update_z(DT ** Z,DT *t,DT * c,DT *columnIsNa) const {

	for (int i=0; i<this->n; i++)
		for (int j=0; j<this->p; j++)
			Z[i][j] -= t[i]*c[j]*columnIsNa[j];
}

void workPls::step_update_y(DT * Y,DT *t,DT d) const {
	for (int i=0; i<this->n; i++)
	{
		Y[i] -= t[i]*d;
	}
}

DT workPls::step_calcul_d(DT * Y,DT *t,DT crosst) const {
	DT d = 0.0;
	DT invcrosst = 1.0 / crosst;
	for (int i=0; i<n; i++)
	{
		d += Y[i]*t[i]*invcrosst;
	}
	return d;
}


void workPls::read(ifstream *in,int & nrem,int **listrem) {
	cout << "Read binary file" << endl ;

	in->read((char *) &n, sizeof(n));
	in->read((char *) &p, sizeof(p));
	in->read((char *) &ncomp, sizeof(ncomp));

	cout << "size population ref="<< n << endl;
	cout << "nb marker used ="<< p << endl;
	cout << "ncomp="<< ncomp << endl;

	this->t = new DT*[ncomp];
	this->u = new DT*[ncomp];
	this->a = new DT*[ncomp];
	this->b = new DT[ncomp];
	this->c = new DT*[ncomp];

	for (int h=0;h<ncomp;h++) {
		//cout << "reading h="<< h << endl;
		this->t[h] = new DT[n];
		in->read((char *) t[h], sizeof(DT)*n);
		this->u[h] = new DT[n];
		in->read((char *) u[h], sizeof(DT)*n);
		this->a[h] = new DT[p];
		in->read((char *) a[h], sizeof(DT)*p);
		this->c[h] = new DT[p];
		in->read((char *) c[h], sizeof(DT)*p);
	}

	in->read((char *) b, sizeof(DT)*ncomp);
	meanZ = new DT[p];
	sigmaZ = new DT[p];
	in->read((char *) meanZ, sizeof(DT)*p);
	in->read((char *) sigmaZ, sizeof(DT)*p);
	in->read((char *) &meanY, sizeof(meanY));
	in->read((char *) &sigmaY, sizeof(sigmaY));
	Y = new DT[n];
	in->read((char *) Y, sizeof(DT)*n);
	in->read((char *) &nrem, sizeof(nrem));
	*listrem = new int[nrem];
	in->read((char *) *listrem, sizeof(int)*nrem);
	cout << "done" << endl ;
}


void workPls::write(ofstream *out,int nrem,int *listrem) const {
	cout << "Write binary file" << endl ;

	out->write((char *) &n, sizeof(n));
	out->write((char *) &p, sizeof(p));
	out->write((char *) &ncomp, sizeof(ncomp));

	for (int h=0;h<ncomp;h++) {
		out->write((char *) t[h], sizeof(DT)*n);
		out->write((char *) u[h], sizeof(DT)*n);
		out->write((char *) a[h], sizeof(DT)*p);
		out->write((char *) c[h], sizeof(DT)*p);
	}

	out->write((char *) b, sizeof(DT)*ncomp);

	out->write((char *) meanZ, sizeof(DT)*p);
	out->write((char *) sigmaZ, sizeof(DT)*p);
	out->write((char *) &meanY, sizeof(meanY));
	out->write((char *) &sigmaY, sizeof(sigmaY));
	out->write((char *) Y, sizeof(DT)*n);
	out->write((char *) &nrem, sizeof(nrem));
	out->write((char *) listrem, sizeof(int)*nrem);
	cout << "done" << endl ;
}


