/*
 *   This file is part of pls4snp.
 *
 *   pls4snp is free software; you can redistribute it and/or modify
 *   it under the terms of the GNU General Public License as published by
 *   the Free Software Foundation; either version 2 of the License, or
 *   (at your option) any later version.
 *
 *   pls4snp is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *   GNU General Public License for more details.
 *
 *   You should have received a copy of the GNU General Public License
 *   along with pls4snp; if not, write to the Free Software
 *   Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
 *
 *   Copyright 2013 Olivier Filangi
 *
 */
/*
 * predict.cpp
 *
 *  Created on: 1 mars 2013
 *      Author: ofilangi
 */

#include <string>
#include <iostream>
#include <fstream>
using namespace std;

#include "common.hpp"
#include "predict.hpp"
#include "convert.hpp"

solPls::solPls() {
	n=0;
	p=0;
	nsol=0;
	yhat=NULL;
	tpred=NULL;
	bhat=NULL;
}

solPls::~solPls() {

	for (int h=0; h<nsol; h++)
	{
		if ( yhat && yhat[h] ) delete[] yhat[h];
		if ( tpred && tpred[h] ) delete[] tpred[h];
		if ( bhat && bhat) delete[] bhat[h];
	}
	if ( yhat ) delete[] yhat ;
	if ( tpred ) delete[] tpred ;
	if ( bhat ) delete[] bhat ;
}

/*
 * Interface to call predict from a programme C/C++/fortran
 *
 * Purpose:
 *
 * Parameters:
 *   [in]  ncomp              : specifies number of latent variable associated with t and a
 *   [in]  n                  : specifies number of row of Y
 *   [in]  p                  : specifies number of SNP
 *   [in]  a[p,ncomp]         : loading vectors associated to X
 *   [in]  c[p,ncomp]         : buffer result from spls subroutine
 *   [in]  b[ncomp]           : loading vectors associated to Y
 *   [in]  meanY              : mean of Y
 *   [in]  sigmY              : standard deviation of Y
 *   [in]  meanZ[p]           : vector of mean of each SNP column
 *   [in]  sigmZ[p]           : vector of standard deviation of each SNP column
 *   [in]  n_pred             : specifies number of row of Z_pred
 *   [in]  Z_pred[n_pred,p]   : incidence matrix to predict
 *   [in]  hcomp              : specifies the number of latente variate to compute yhat,tpred,bhat
 *   [out] yhat[n_pred,hcomp] : Y prediction
 *   [out] tpred[n_pred,hcomp]:
 *   [out] bhat[p,hcomp]      :
 */

extern "C" void predict_(int *n, int *p, int *ncomp, DT * a, DT* c, DT * b,DT * meanY, DT * sigmaY, DT * meanZ, DT* sigmaZ,int * n_pred
		,DT * Z_pred,int * hcomp, DT * yhat,DT * tpred,DT * bhat) {

	solPls sol;
	workPls work;
	sol.n = *n;
	sol.p = *p;
	work.n = *n;
	work.p = *p;
	work.ncomp = *ncomp;

	int hPredict = *hcomp ;
	int * listOk = new int[sol.p];
	/* manage all snp */
	for (int i=0;i<sol.p;i++)
		listOk[i] = i;

	work.a = new DT*[*ncomp];
	work.c = new DT*[*ncomp];
	work.b = new DT[*ncomp];

	for (int h=0;h<*ncomp;h++) {
		work.a[h] = new DT[*p];
		work.c[h] = new DT[*p];
		for (int i=0;i<sol.p;i++) {
			work.a[h][i] = a[i+h*(sol.p)];
			work.c[h][i] = c[i+h*(sol.p)];
		}
		work.b[h] = b[h];
	}

	work.meanY = *meanY;
	work.sigmaY = *sigmaY;

	work.meanZ = meanZ;
	work.sigmaZ = sigmaZ;

	DT ** X_pred = new DT*[*n_pred];

	for (int i=0;i<*n_pred;i++) {
		X_pred[i] = new DT[*p];
		for (int j=0;j<*p;j++) {
			X_pred[i][j] = Z_pred[i+j*(*n_pred)];
		}
	}

	sol.predict(work,X_pred,*n_pred,sol.p,listOk,hPredict);

	/* results */

	for (int h=0;h<hPredict;h++) {
		for (int i=0;i<sol.n;i++) {
			yhat[i+h*(sol.n)] = sol.yhat[h][i];
			tpred[i+h*(sol.n)] = sol.tpred[h][i];
		}
		for (int i=0;i<sol.p;i++) {
			bhat[i+h*(sol.n)] = sol.bhat[h][i];
		}
	}

	for (int i=0;i<sol.n;i++)
		delete[] X_pred[i];

	for (int h=0;h<*ncomp;h++) {
		delete[] work.a[h];
		delete[] work.c[h];
	}
	delete[] work.a;
	delete[] work.c;
	delete[] work.b;

	work.a = NULL;
	work.c = NULL;
	work.b = NULL;
	work.meanZ = NULL;
	work.sigmaZ = NULL;

	delete[] X_pred;
	delete [] listOk;
}

/*
 Fonction pour predire les phenotypes des n_pred individus genotypés (X_pred)
 */
void solPls::predict(workPls & work,DT ** X_pred, int n_pred,int pAll,int * listOk, int hPredict)
{
	if ( hPredict < 0 && hPredict != -1 ) {
		cerr << "h can not be negative. h="<<  hPredict << endl ;
		exit(1);
	}

	if ( hPredict > work.ncomp ) {
		cerr << "can not predict with H="<< hPredict << " ncomp="<< work.ncomp << endl ;
		exit(1);
	}

	DT * W = new DT[work.p*work.ncomp];
	if ( hPredict == -1 ) {
		this->nsol = work.ncomp;
		this->bhat = new DT*[work.ncomp];
		this->yhat = new DT*[work.ncomp];
		this->tpred = new DT*[work.ncomp];
	} else {
		this->nsol = hPredict ;
		this->bhat = new DT*[hPredict];
		this->yhat = new DT*[hPredict];
		this->tpred = new DT*[hPredict];
	}

	this->n = n_pred;
	this->p = work.p;


	DT * buf_a = new DT[work.ncomp];
	DT * workInv = new DT[work.ncomp*2];

	/* Note OFI : on sort l'inversion de matrice pour trouver les vecteurs loadinds */

	/*
					   R: W = a[, 1:h] %*% solve(t(c[, 1:h]) %*% a[, 1:h])
	 */
	int dimh=work.ncomp;
	DT at[dimh*dimh] ;//[dimh*dimh];//column major order ===> offset = row + column*NUMROWS
	/* t(c[, 1:h]) %*% a[, 1:h] */

	for (int combih=0; combih<dimh*dimh; combih++)
	{
		int ih = combih / dimh ;
		int ih2 =  combih % dimh ;
		DT v = 0 ;
		for (int j=0; j<p; j++)
		{
			v += work.c[ih][j]*work.a[ih2][j];
		}
		at[ih+ih2*dimh]=v;
	}

	int ipiv[dimh];
	int info;
	/* LU factorization - Lapack routine */
	getrf( &dimh, &dimh, at, &dimh, ipiv, &info );
	if ( info == 0 )
	{
		//  cout << "LU decomposition (dgetrf) ok"<< endl ;
	}
	else if (info < 0 )
	{
		cout << "LU decomposition (dgetrf) illegal value"<< endl ;
	}
	else
	{
		cout << "LU decomposition (dgetrf) failed - info="<< info << endl;
	}

	//  DGETRI computes the inverse of a matrix using the LU factorization
	//  computed by DGETRF.
	int lwork=dimh*2;

	getri(&dimh, at, &dimh, ipiv, workInv, &lwork,&info);

	if (info != 0 )
	{
		cerr << "Matrix inversion failed!" << endl ;
		return;
	}

	for (int i=0; i<p; i++)
	{
		/* optimisation - acces memory linear */
		for (int k=0; k<dimh; k++)
			buf_a[k] = work.a[k][i];



		for(int h=0;h<work.ncomp;h++) {
			DT v=0;
			for (int k=0; k<dimh; k++)
			{
				v += at[k+h*dimh]*buf_a[k];
			}
			W[i+h*work.p]= v ;
		}
	}
	//	for (int i=0;i<10;i++) {
	//		for(int h=0;h<work.ncomp;h++) {
	//			cout << W[i+h*work.p] << " ";
	//		}
	//		cout << endl;
	//	}
	int hsol = 0;
	cout << "H=";
	for (int h=0; h<work.ncomp; h++)
	{
		if ( (hPredict != -1) && h >= hPredict ) break;

		dimh = h+1;

		this->bhat[hsol] = new DT[pAll]; /* on met 0.0 pour les marqueurs qui ne sont pas pris dans l analyse*/
		this->yhat[hsol] = new DT[n_pred];
		this->tpred[hsol] = new DT[n_pred];

		/*
          R:  B = W %*% drop(t(b[, 1:h]))
		 */
		for (int i=0; i<pAll; i++)
		{
			this->bhat[hsol][i]=0;
		}
		for (int ih=0;ih<dimh;ih++) {
			for (int i=0; i<p; i++)
				this->bhat[hsol][listOk[i]] += W[i+ih*work.p]*work.b[ih];
		}
		//test +=   W[0+h*work.p]*work.b[h] ;
		//	cout << this->bhat[hsol][0] << endl ;

		DT intercept=0;
		for (int i=0; i<p; i++)
		{
			// B = scale(B, center = FALSE, scale = 1 / sigma.Y)
			// and
			// B = as.matrix(scale(t(B), center = FALSE, scale = sigma.X))
			// and
			// intercept = -scale(B, center = FALSE, scale = 1 / means.X)
			this->bhat[hsol][listOk[i]] *= work.sigmaY / work.sigmaZ[i];
			/* TODO ofilangi : A verifier car je n arrive pas a le valider avec R (valeur Nan en retour)*/
			intercept -= this->bhat[hsol][listOk[i]]*work.meanZ[i];
		}

		intercept += work.meanY;
		DT vyhat = 0;
		DT vtpred = 0;

		for (int i=0;i<n_pred;i++) {
			vyhat = 0;
			vtpred = 0;

			for (int j=0;j<p;j++) {
				vyhat += this->bhat[hsol][listOk[j]]*X_pred[i][j];
				vtpred += ((X_pred[i][j] - work.meanZ[j])/work.sigmaZ[j])*W[j+h*work.p];
			}

			this->tpred[hsol][i] = vtpred;
			this->yhat[hsol][i] = vyhat + intercept;
			//cout << vtpred << endl;
		}
		cout << h+1 << " ";
		cout.flush();
		hsol++;
	}
	cout << endl ;

	delete[] W;
	delete[] buf_a;
	delete[] workInv;
}

void solPls::write_sol(const dataGen & data,string nameOut,int hselect) const {

	std::ofstream tpredstream;
	tpredstream.open(nameOut.c_str());

	for (map<string,int>::const_iterator it = data.genoId.begin();it != data.genoId.end();it++) {
		tpredstream << (*it).first << " "; /* ID */
		if ( hselect >= 0 && hselect<nsol ) {
			tpredstream << yhat[hselect][(*it).second];
		} else {
			for (int h=0;h<nsol;h++) {
				tpredstream << yhat[h][(*it).second]<< " ";
			}
		}
		tpredstream << endl ;
	}
	tpredstream.close();
	cout << "Write file ["<< nameOut << "] phenotypes prediction." << endl ;
}
